@@ -27,6 +27,10 @@
 #include "util.h"
 #include "system_area.h"
 
+#ifdef Libisofs_with_checksumS
+#include "md5.h"
+#endif
+
 #include <stdlib.h>
 #include <time.h>
 #include <string.h>
@@ -45,19 +49,38 @@ void ecma119_image_free(Ecma119Image *t)
 {
     size_t i;
 
-    ecma119_node_free(t->root);
-    iso_image_unref(t->image);
-    iso_rbtree_destroy(t->files, iso_file_src_free);
-    iso_ring_buffer_free(t->buffer);
+    if (t == NULL)
+        return;
+    if (t->root != NULL)
+        ecma119_node_free(t->root);
+    if (t->image != NULL)
+        iso_image_unref(t->image);
+    if (t->files != NULL)
+        iso_rbtree_destroy(t->files, iso_file_src_free);
+    if (t->buffer != NULL)
+        iso_ring_buffer_free(t->buffer);
 
     for (i = 0; i < t->nwriters; ++i) {
         IsoImageWriter *writer = t->writers[i];
         writer->free_data(writer);
         free(writer);
     }
-    free(t->input_charset);
-    free(t->output_charset);
-    free(t->writers);
+    if (t->input_charset != NULL)
+        free(t->input_charset);
+    if (t->output_charset != NULL)
+        free(t->output_charset);
+
+#ifdef Libisofs_with_checksumS
+    if (t->checksum_ctx != NULL) { /* dispose checksum context */
+        char md5[16];
+        iso_md5_end(&(t->checksum_ctx), md5);
+    }
+    if (t->checksum_buffer != NULL)
+        free(t->checksum_buffer);
+#endif
+
+    if (t->writers != NULL)
+        free(t->writers);
     free(t);
 }
 
@@ -227,6 +250,16 @@ int ecma119_writer_compute_data_blocks(IsoImageWriter *writer)
     target->curblock += DIV_UP(path_table_size, BLOCK_SIZE);
     target->path_table_size = path_table_size;
 
+#ifdef Libisofs_with_checksumS
+
+    if (target->md5_session_checksum) {
+        /* Account for tree checksum tag */
+        target->checksum_tree_tag_pos = target->curblock;
+        target->curblock++;
+    }
+
+#endif /* Libisofs_with_checksumS */
+
     return ISO_SUCCESS;
 }
 
@@ -294,7 +327,7 @@ void write_one_dir_record(Ecma119Image *t, Ecma119Node *node, int file_id,
     rec->len_dr[0] = len_dr + (info != NULL ? info->suf_len : 0);
     iso_bb(rec->block, block, 4);
     iso_bb(rec->length, len, 4);
-    if(t->dir_rec_mtime) {
+    if (t->dir_rec_mtime) {
         iso= node->node;
         iso_datetime_7(rec->recording_time,
                        t->replace_timestamps ? t->timestamp : iso->mtime,
@@ -660,6 +693,17 @@ int ecma119_writer_write_data(IsoImageWriter *writer)
 
     /* and write the path tables */
     ret = write_path_tables(t);
+    if (ret < 0)
+        return ret;
+
+#ifdef Libisofs_with_checksumS
+
+    if (t->md5_session_checksum) {
+        /* Write tree checksum tag */
+        ret = iso_md5_write_tag(t, 3);
+    }
+
+#endif /* Libisofs_with_checksumS */
 
     return ret;
 }
@@ -783,6 +827,19 @@ int pad_writer_create(Ecma119Image *target)
     return ISO_SUCCESS;
 }
 
+static
+int transplant_checksum_buffer(Ecma119Image *target, int flag)
+{
+    /* Transplant checksum buffer from Ecma119Image to IsoImage */
+    iso_image_set_checksums(target->image, target->checksum_buffer,
+                            target->checksum_range_start,
+                            target->checksum_array_pos,
+                            target->checksum_idx_counter + 2, 0);
+    target->checksum_buffer = NULL;
+    target->checksum_idx_counter = 0;
+    return 1;
+}
+
 static
 void *write_function(void *arg)
 {
@@ -835,6 +892,18 @@ void *write_function(void *arg)
         }
     }
 
+#ifdef Libisofs_with_checksumS
+
+    /* Write superblock checksum tag */
+    if (target->md5_session_checksum && target->checksum_ctx != NULL) {
+        res = iso_md5_write_tag(target, 2);
+        if (res < 0)
+            goto write_error;
+    }
+
+#endif /* Libisofs_with_checksumS */
+
+
     /* write data for each writer */
     for (i = 0; i < target->nwriters; ++i) {
         writer = target->writers[i];
@@ -844,6 +913,13 @@ void *write_function(void *arg)
         }
     }
 
+#ifdef Libisofs_with_checksumS
+
+    /* Transplant checksum buffer from Ecma119Image to IsoImage */
+    transplant_checksum_buffer(target, 0);
+
+#endif
+
     iso_ring_buffer_writer_close(target->buffer, 0);
     pthread_exit(NULL);
 
@@ -857,13 +933,122 @@ void *write_function(void *arg)
                    "Image write error");
     }
     iso_ring_buffer_writer_close(target->buffer, 1);
+
+#ifdef Libisofs_with_checksumS
+
+    /* Transplant checksum buffer from Ecma119Image to IsoImage */
+    transplant_checksum_buffer(target, 0);
+
+#endif
+
     pthread_exit(NULL);
 }
 
+
+#ifdef Libisofs_with_checksumS
+
+
+static
+int checksum_prepare_image(IsoImage *src, int flag)
+{
+    int ret;
+
+    /* Set provisory value of isofs.ca with
+       4 byte LBA, 4 byte count, size 16, name MD5 */
+    ret = iso_root_set_isofsca((IsoNode *) src->root, 0, 0, 0, 16, "MD5", 0);
+    if (ret < 0)
+        return ret;
+    return ISO_SUCCESS;
+}
+
+
+/*
+  @flag bit0= recursion
+*/
+static
+int checksum_prepare_nodes(Ecma119Image *target, IsoNode *node, int flag)
+{
+    IsoNode *pos;
+    IsoFile *file;
+    IsoImage *img;
+    int ret, i, no_md5 = 0, has_xinfo = 0;
+    size_t value_length;
+    unsigned int idx = 0;
+    char *value= NULL;
+    void *xipt = NULL;
+    static char *cx_names = "isofs.cx";
+    static size_t cx_value_lengths[1] = {0};
+    char *cx_valuept = "";
+
+    img= target->image;
+
+    if (node->type == LIBISO_FILE) {
+        file = (IsoFile *) node;
+        if (file->from_old_session && target->appendable) {
+            /* Save MD5 data of files from old image which will not
+               be copied and have an MD5 recorded in the old image. */
+            has_xinfo = iso_node_get_xinfo(node, checksum_md5_xinfo_func,
+                                           &xipt);
+            if (has_xinfo <= 0) {
+                ret = iso_node_lookup_attr(node, "isofs.cx", &value_length,
+                                           &value, 0);
+            }
+            if (has_xinfo > 0) {
+                /* xinfo MD5 overrides everything else unless data get copied
+                   and checksummed during that copying
+                 */;
+            } else if (ret == 1 && img->checksum_array == NULL) {
+                /* No checksum array loaded. Delete "isofs.cx" */
+                iso_node_set_attrs(node, (size_t) 1,
+                              &cx_names, cx_value_lengths, &cx_valuept, 4 | 8);
+                no_md5 = 1;
+            } else if (ret == 1 && value_length == 4) {
+                for (i = 0; i < 4; i++)
+                    idx = (idx << 8) | ((unsigned char *) value)[i];
+                if (idx > 0 && idx < 0x8000000) {
+                    /* xipt is an int disguised as void pointer */
+                    for (i = 0; i < 4; i++)
+                        ((char *) &xipt)[i] = value[i];
+                    ret = iso_node_add_xinfo(node, checksum_cx_xinfo_func,
+                                             xipt);
+                    if (ret < 0)
+                        return ret;
+                } else
+                    no_md5 = 1;
+            } else {
+                no_md5 = 1;
+            }
+            if (value != NULL) {
+                free(value);
+                value= NULL;
+            }
+        }
+        /* Equip nodes with provisory isofs.cx numbers: 4 byte, all 0.
+           Omit those from old image which will not be copied and have no MD5.
+         */
+        if (!no_md5) {
+            ret = iso_file_set_isofscx(file, (unsigned int) 0, 0);
+            if (ret < 0)
+                return ret;
+        }
+    } else if (node->type == LIBISO_DIR) {
+        for (pos = ((IsoDir *) node)->children; pos != NULL; pos = pos->next) {
+            ret = checksum_prepare_nodes(target, pos, 1);
+            if (ret < 0)
+                return ret;
+        }
+    }
+    return ISO_SUCCESS;
+}
+
+
+#endif /* Libisofs_with_checksumS */
+
+
 static
 int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
 {
-    int ret, i, voldesc_size, nwriters;
+    int ret, i, voldesc_size, nwriters, image_checksums_mad = 0;
     Ecma119Image *target;
     int el_torito_writer_index = -1, file_src_writer_index= -1;
 
@@ -876,8 +1061,7 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
     /* create the tree for file caching */
     ret = iso_rbtree_new(iso_file_src_cmp, &(target->files));
     if (ret < 0) {
-        free(target);
-        return ret;
+        goto target_cleanup;
     }
 
     target->image = src;
@@ -890,11 +1074,6 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
     target->hardlinks = opts->hardlinks;
     target->aaip = opts->aaip;
     target->always_gmt = opts->always_gmt;
-
-#ifndef Libisofs_hardlink_matcheR
-    target->ino = 0;
-#endif
-
     target->omit_version_numbers = opts->omit_version_numbers
                                  | opts->max_37_char_filenames;
     target->allow_deep_paths = opts->allow_deep_paths;
@@ -935,9 +1114,8 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
 
     target->input_charset = strdup(iso_get_local_charset(0));
     if (target->input_charset == NULL) {
-        iso_image_unref(src);
-        free(target);
-        return ISO_OUT_OF_MEM;
+        ret = ISO_OUT_OF_MEM;
+        goto target_cleanup;
     }
 
     if (opts->output_charset != NULL) {
@@ -946,11 +1124,29 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
         target->output_charset = strdup(target->input_charset);
     }
     if (target->output_charset == NULL) {
-        iso_image_unref(src);
-        free(target);
-        return ISO_OUT_OF_MEM;
+        ret = ISO_OUT_OF_MEM;
+        goto target_cleanup;
     }
 
+#ifdef Libisofs_with_checksumS
+
+    target->md5_file_checksums = opts->md5_file_checksums;
+    target->md5_session_checksum = opts->md5_session_checksum;
+    target->checksum_idx_counter = 0;
+    target->checksum_ctx = NULL;
+    target->checksum_counter = 0;
+    target->checksum_rlsb_tag_pos = 0;
+    target->checksum_sb_tag_pos = 0;
+    target->checksum_tree_tag_pos = 0;
+    target->checksum_tag_pos = 0;
+    target->checksum_buffer = NULL;
+    target->checksum_array_pos = 0;
+    target->checksum_range_start = 0;
+    target->checksum_range_size = 0;
+    target->opts_overwrite = 0;
+
+#endif
+
     /*
      * 2. Based on those options, create needed writers: iso, joliet...
      * Each writer inits its structures and stores needed info into
@@ -974,11 +1170,32 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
         nwriters++;
     }
 
+
+#ifdef Libisofs_with_checksumS
+
+    if ((target->md5_file_checksums & 1) || target->md5_session_checksum) {
+        nwriters++;
+        image_checksums_mad = 1; /* from here on the loaded checksums are
+                                    not consistent with isofs.cx any more.
+                                  */
+        ret = checksum_prepare_image(src, 0);
+        if (ret < 0)
+            goto target_cleanup;
+        if (target->appendable) {
+            ret = checksum_prepare_nodes(target, (IsoNode *) src->root, 0);
+            if (ret < 0)
+                goto target_cleanup;
+        }
+        target->checksum_idx_counter = 0;
+    }
+
+#endif /* Libisofs_with_checksumS */
+
+
     target->writers = malloc(nwriters * sizeof(void*));
     if (target->writers == NULL) {
-        iso_image_unref(src);
-        free(target);
-        return ISO_OUT_OF_MEM;
+        ret = ISO_OUT_OF_MEM;
+        goto target_cleanup;
     }
 
     /* create writer for ECMA-119 structure */
@@ -1019,7 +1236,7 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
 
     /*
      * Create the writer for possible padding to ensure that in case of image
-     * growing we can safety overwrite the first 64 KiB of image.
+     * growing we can safely overwrite the first 64 KiB of image.
      */
     ret = pad_writer_create(target);
     if (ret < 0) {
@@ -1033,6 +1250,18 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
     }
     file_src_writer_index = target->nwriters - 1;
 
+
+#ifdef Libisofs_with_checksumS
+
+    if ((target->md5_file_checksums & 1) || target->md5_session_checksum) {
+        ret = checksum_writer_create(target);
+        if (ret < 0)
+            goto target_cleanup;
+    }
+
+#endif /* Libisofs_with_checksumS */
+
+
     /*
      * 3.
      * Call compute_data_blocks() in each Writer.
@@ -1078,7 +1307,6 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
 
     /* check if we need to provide a copy of volume descriptors */
     if (opts->overwrite) {
-
         /*
          * Get a copy of the volume descriptors to be written in a DVD+RW
          * disc
@@ -1131,6 +1359,28 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
         vol->vol_desc_type[0] = 255;
         memcpy(vol->std_identifier, "CD001", 5);
         vol->vol_desc_version[0] = 1;
+
+#ifdef Libisofs_with_checksumS
+
+        /* Write relocated superblock checksum tag */
+        if (target->md5_session_checksum) {
+            target->checksum_rlsb_tag_pos = voldesc_size / BLOCK_SIZE + 16 + 1;
+            if (target->checksum_rlsb_tag_pos < 32) {
+                ret = iso_md5_start(&(target->checksum_ctx));
+                if (ret < 0)
+                    goto target_cleanup;
+                target->opts_overwrite = (char *) opts->overwrite;
+                iso_md5_compute(target->checksum_ctx, target->opts_overwrite,
+                                target->checksum_rlsb_tag_pos * 2048);
+                ret = iso_md5_write_tag(target, 4);
+                target->opts_overwrite = NULL; /* opts might not persist */
+                if (ret < 0)
+                    goto target_cleanup;
+            }
+        }
+
+#endif /* Libisofs_with_checksumS */
+
     }
 
     /*
@@ -1140,7 +1390,24 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
     target->vol_space_size = target->curblock - target->ms_block;
     target->total_size = (off_t) target->vol_space_size * BLOCK_SIZE;
 
-    /* 4. Create and start writting thread */
+
+    /* 4. Create and start writing thread */
+
+#ifdef Libisofs_with_checksumS
+    if (target->md5_session_checksum) {
+        /* After any fake writes are done: Initialize image checksum context */
+        if (target->checksum_ctx != NULL)
+            iso_md5_end(&(target->checksum_ctx), target->image_md5);
+        ret = iso_md5_start(&(target->checksum_ctx));
+        if (ret < 0)
+            goto target_cleanup;
+    }
+    /* Dispose old image checksum buffer. The one of target is supposed to
+       get attached at the end of write_function(). */
+    iso_image_free_checksums(target->image, 0);
+    image_checksums_mad = 0;
+
+#endif /* Libisofs_with_checksumS */
 
     /* ensure the thread is created joinable */
     pthread_attr_init(&(target->th_attr));
@@ -1166,6 +1433,14 @@ int ecma119_image_new(IsoImage *src, IsoWriteOpts *opts, Ecma119Image **img)
     return ISO_SUCCESS;
 
     target_cleanup: ;
+
+#ifdef Libisofs_with_checksumS
+
+    if(image_checksums_mad) /* No checksums is better than mad checksums */
+      iso_image_free_checksums(target->image, 0);
+
+#endif /* Libisofs_with_checksumS */
+
     ecma119_image_free(target);
     return ret;
 }
@@ -1306,6 +1581,16 @@ int iso_write(Ecma119Image *target, void *buf, size_t count)
         return ISO_CANCELED;
     }
 
+#ifdef Libisofs_with_checksumS
+
+    if (target->checksum_ctx != NULL) {
+        /* Add to image checksum */
+        target->checksum_counter += count;
+        iso_md5_compute(target->checksum_ctx, (char *) buf, (int) count);
+    }
+
+#endif /* Libisofs_with_checksumS */
+
     /* total size is 0 when writing the overwrite buffer */
     if (ret > 0 && (target->total_size != (off_t) 0)){
         unsigned int kbw, kbt;
@@ -1568,6 +1853,19 @@ int iso_write_opts_set_sort_files(IsoWriteOpts *opts, int sort)
     return ISO_SUCCESS;
 }
 
+int iso_write_opts_set_record_md5(IsoWriteOpts *opts, int session, int files)
+{
+
+#ifdef Libisofs_with_checksumS
+
+    opts->md5_session_checksum = session & 1;
+    opts->md5_file_checksums = files & 3;
+
+#endif /* Libisofs_with_checksumS */
+
+    return ISO_SUCCESS;
+}
+
 int iso_write_opts_set_replace_mode(IsoWriteOpts *opts, int dir_mode,
                                     int file_mode, int uid, int gid)
 {
@@ -1710,7 +2008,7 @@ int iso_write_opts_set_fifo_size(IsoWriteOpts *opts, size_t fifo_size)
 int iso_write_opts_get_data_start(IsoWriteOpts *opts, uint32_t *data_start,
                                   int flag)
 {
-    if(opts->data_start_lba == 0)
+    if (opts->data_start_lba == 0)
 	return ISO_ERROR;
     *data_start = opts->data_start_lba;
     return ISO_SUCCESS;
