@@ -147,6 +147,30 @@ struct iso_write_opts {
      */
     unsigned int dir_rec_mtime :1;
 
+#ifdef Libisofs_with_checksumS
+
+    /**
+     * Compute MD5 checksum for the whole session and record it as index 0 of
+     * the checksum blocks after the data area of the session. The layout and
+     * position of these blocks will be recorded in xattr "isofs.ca" of the
+     * root node. See see also API call iso_image_get_session_md5().
+     */
+    unsigned int md5_session_checksum :1;
+
+    /**
+     * Compute MD5 checksums for IsoFile objects and write them to blocks
+     * after the data area of the session. The layout and position of these
+     * blocks will be recorded in xattr "isofs.ca" of the root node.
+     * The indice of the MD5 sums will be recorded with the IsoFile directory
+     * entries as xattr "isofs.cx". See also API call iso_file_get_md5().
+     * bit0= compute individual checksums
+     * bit1= pre-compute checksum and compare it with actual one.
+     *       Raise MISHAP if mismatch.
+     */
+    unsigned int md5_file_checksums :2;
+
+#endif /* Libisofs_with_checksumS */
+
     /** If files should be sorted based on their weight. */
     unsigned int sort_files :1;
 
@@ -312,6 +336,13 @@ struct ecma119_image
     /* Store in ECMA-119 timestamp mtime of source */
     unsigned int dir_rec_mtime :1;
 
+#ifdef Libisofs_with_checksumS
+
+    unsigned int md5_session_checksum :1;
+    unsigned int md5_file_checksums :2;
+
+#endif /* Libisofs_with_checksumS */
+
     /*
      * Mode replace. If one of these flags is set, the correspodent values are
      * replaced with values below.
@@ -333,18 +364,6 @@ struct ecma119_image
      */
     int sort_files;
 
-
-#ifndef Libisofs_hardlink_matcheR
-
-    /* ts A90508 : <<< this is on its way out */
-    /**
-     * In the CD, each file must have an unique inode number. So each
-     * time we add a new file, this is incremented.
-     */
-    ino_t ino;
-
-#endif /* ! Libisofs_hardlink_matcheR */
-
     char *input_charset;
     char *output_charset;
 
@@ -421,6 +440,32 @@ struct ecma119_image
     /* tree of files sources */
     IsoRBTree *files;
 
+#ifdef Libisofs_with_checksumS
+
+    unsigned int checksum_idx_counter;
+    void *checksum_ctx;
+    off_t checksum_counter;
+    uint32_t checksum_rlsb_tag_pos;
+    uint32_t checksum_sb_tag_pos;
+    uint32_t checksum_tree_tag_pos;
+    uint32_t checksum_tag_pos;
+    char image_md5[16];
+    char *checksum_buffer;
+    uint32_t checksum_array_pos;
+    uint32_t checksum_range_start;
+    uint32_t checksum_range_size;
+
+    char *opts_overwrite; /* Points to IsoWriteOpts->overwrite.
+                             Use only underneath ecma119_image_new()
+                             and if not NULL*/
+
+    /* ??? Is there a reason why we copy lots of items from IsoWriteOpts
+           rather than taking ownership of the IsoWriteOpts object which
+           is submitted with ecma119_image_new() ?
+     */
+
+#endif /* Libisofs_with_checksumS */
+
     /* Buffer for communication between burn_source and writer thread */
     IsoRingBuffer *buffer;
 
