@@ -13,6 +13,7 @@
 #include "stream.h"
 #include "aaip_0_2.h"
 #include "messages.h"
+#include "util.h"
 
 
 #include <stdlib.h>
@@ -1309,13 +1310,9 @@ int iso_node_new_symlink(char *name, char *dest, IsoSymlink **link)
     new->node.name = name;
     new->dest = dest;
     new->node.mode = S_IFLNK;
-
-#ifdef Libisofs_hardlink_matcheR
     new->fs_id = 0;
     new->st_dev = 0;
     new->st_ino = 0;
-#endif
-
     *link = new;
     return ISO_SUCCESS;
 }
@@ -1347,13 +1344,9 @@ int iso_node_new_special(char *name, mode_t mode, dev_t dev,
 
     new->node.mode = mode;
     new->dev = dev;
-
-#ifdef Libisofs_hardlink_matcheR
     new->fs_id = 0;
     new->st_dev = 0;
     new->st_ino = 0;
-#endif
-
     *special = new;
     return ISO_SUCCESS;
 }
@@ -2300,8 +2293,6 @@ int iso_node_get_id(IsoNode *node, unsigned int *fs_id, dev_t *dev_id,
         }
         return 1;
 
-#ifdef Libisofs_hardlink_matcheR
-
     } else if (node->type == LIBISO_SYMLINK) {
         symlink = (IsoSymlink *) node;
         if (symlink->fs_id != ISO_IMAGE_FS_ID && (flag & 1)) {
@@ -2324,8 +2315,6 @@ int iso_node_get_id(IsoNode *node, unsigned int *fs_id, dev_t *dev_id,
         *ino_id = special->st_ino;
         return 1;
 
-#endif
-
     }
 
     ret = 0;
@@ -2380,8 +2369,6 @@ int iso_node_set_ino(IsoNode *node, ino_t ino, int flag)
         if (ret < 0 || ret == 1)
             return ret;
 
-#ifdef Libisofs_hardlink_matcheR
-
     } else if (node->type == LIBISO_SYMLINK) {
         symlink = (IsoSymlink *) node;
         if (symlink->fs_id == ISO_IMAGE_FS_ID) {
@@ -2396,8 +2383,6 @@ int iso_node_set_ino(IsoNode *node, ino_t ino, int flag)
             return 1;
         }
 
-#endif
-
     }
     ret = iso_node_set_ino_xinfo(node, ino, 0);
     if (ret < 0)
@@ -2472,8 +2457,6 @@ int iso_node_cmp_flag(IsoNode *n1, IsoNode *n2, int flag)
             return ret1;
         goto inode_match;
 
-#ifdef Libisofs_hardlink_matcheR
-
     } else if (n1->type == LIBISO_SYMLINK) {
 
         l1 = (IsoSymlink *) n1;
@@ -2496,8 +2479,6 @@ int iso_node_cmp_flag(IsoNode *n1, IsoNode *n2, int flag)
         dev_id2 = s2->st_dev;
         ino_id2 = s2->st_ino;
 
-#endif /* Libisofs_hardlink_matcheR */
-
     } else {
         return (n1 < n2 ? -1 : 1); /* case n1 == n2 is handled above */
     }
@@ -2575,3 +2556,186 @@ int iso_node_cmp_ino(IsoNode *n1, IsoNode *n2, int flag)
 {
     return iso_node_cmp_flag(n1, n2, 1);
 }
+
+
+int iso_file_set_isofscx(IsoFile *file, unsigned int checksum_index,
+                         int flag)
+{
+    static char *names = "isofs.cx";
+    static size_t value_lengths[1] = {4};
+    unsigned char value[4];
+    char *valuept;
+    int i, ret;
+
+    for(i = 0; i < 4; i++)
+        value[3 - i] = (checksum_index >> (8 * i)) & 0xff;
+    valuept= (char *) value;
+    ret = iso_node_set_attrs((IsoNode *) file, (size_t) 1,
+                             &names, value_lengths, &valuept, 2 | 8);
+    return ret;
+}
+
+
+int iso_root_set_isofsca(IsoNode *node, uint32_t start_lba, uint32_t end_lba,
+                         uint32_t count, uint32_t size, char *typetext,
+                         int flag)
+{
+    char buffer[5 + 5 + 5 + 2 + 81], *wpt = buffer, *valuept = buffer;
+    int result_len, ret;
+    static char *names = "isofs.ca";
+    static size_t value_lengths[1];
+
+    /* Set value of isofs.ca with
+       4 byte START, 4 byte END, 4 byte COUNT, SIZE = 16,  MD5 */
+    iso_util_encode_len_bytes(start_lba, wpt, 4, &result_len, 0);
+    wpt += result_len;
+    iso_util_encode_len_bytes(end_lba, wpt, 4, &result_len, 0);
+    wpt += result_len;
+    iso_util_encode_len_bytes(count, wpt, 4, &result_len, 0);
+    wpt += result_len;
+    iso_util_encode_len_bytes(size, wpt, 1, &result_len, 0);
+    wpt += result_len;
+    strncpy(wpt, typetext, 80);
+    if (strlen(typetext) > 80)
+        wpt += 80;
+    else
+        wpt += strlen(typetext);
+    value_lengths[0] = wpt - buffer;
+    ret = iso_node_set_attrs(node, (size_t) 1,
+                             &names, value_lengths, &valuept, 2 | 8);
+    return ret;
+}
+
+
+int iso_root_get_isofsca(IsoNode *node, uint32_t *start_lba, uint32_t *end_lba,
+                         uint32_t *count, uint32_t *size, char typetext[81],
+                         int flag)
+{
+    int ret, len;
+    size_t value_len;
+    char *value = NULL, *rpt;
+
+    ret = iso_node_lookup_attr(node, "isofs.ca", &value_len, &value, 0);
+    if (ret <= 0)
+        goto ex;
+
+    /* Parse value of isofs.ca with
+       4 byte START, 4 byte END, 4 byte COUNT, SIZE = 16,  MD5 */
+    rpt = value;
+    iso_util_decode_len_bytes(start_lba, rpt, &len,
+                              value_len - (rpt - value), 0);
+    rpt += len + 1;
+    iso_util_decode_len_bytes(end_lba, rpt, &len,
+                              value_len - (rpt - value), 0);
+    rpt += len + 1;
+    iso_util_decode_len_bytes(count, rpt, &len,
+                              value_len - (rpt - value), 0);
+    rpt += len + 1;
+    iso_util_decode_len_bytes(size, rpt, &len,
+                              value_len - (rpt - value), 0);
+    rpt += len + 1;
+    len = value_len - (rpt - value);
+    if (len > 80)
+        len = 80;
+    memcpy(typetext, rpt, len);
+    typetext[len] = 0;
+
+    ret= ISO_SUCCESS;
+ex:;
+    if (value != NULL)
+        free(value);
+    return ret;
+}
+
+
+/* API */
+int iso_file_get_md5(IsoImage *image, IsoFile *file, char md5[16], int flag)
+{
+
+#ifdef Libisofs_with_checksumS
+
+    int ret, i;
+    size_t value_len;
+    char *value = NULL;
+    uint32_t idx = 0;
+    void *xipt;
+
+    /* xinfo MD5 overrides everything else */
+    ret = iso_node_get_xinfo((IsoNode *) file, checksum_md5_xinfo_func, &xipt);
+    if (ret == 1) {
+        memcpy(md5, (char *) xipt, 16);
+        return 1;
+    }
+
+    if (image->checksum_array == NULL)
+        return 0;
+    ret = iso_node_lookup_attr((IsoNode *) file, "isofs.cx",
+                               &value_len, &value, 0);
+    if (ret <= 0)
+        goto ex;
+    if (value_len > 4) {
+        ret = 0;
+        goto ex;
+    }
+    for (i = 0; i < value_len; i++)
+        idx = (idx << 8) | ((unsigned char *) value)[i];
+    if (idx == 0 || idx > image->checksum_idx_count - 1) {
+                                       /* (last index is not MD5 of a file) */
+        ret = 0;
+        goto ex;
+    }
+    if (!(flag & 1)) {
+        memcpy(md5, image->checksum_array + ((size_t) 16) * ((size_t) idx),
+               16);
+    }
+    ret = 1;
+ex:;
+    if (value != NULL)
+        free(value);
+    return ret;
+
+#else
+
+    return 0;
+
+#endif /* ! Libisofs_with_checksumS */
+
+}
+
+
+/* API */
+int iso_file_make_md5(IsoFile *file, int flag)
+{
+
+#ifdef Libisofs_with_checksumS
+    int ret, dig = 0;
+    char *md5 = NULL;
+
+    if (file->from_old_session)
+        dig = 1;
+    md5= calloc(16, 1);
+    ret = iso_stream_make_md5(file->stream, md5, dig);
+    if (ret < 0)
+        goto ex;
+    iso_node_remove_xinfo((IsoNode *) file, checksum_md5_xinfo_func);
+    ret = iso_node_add_xinfo((IsoNode *) file, checksum_md5_xinfo_func, md5);
+    if (ret == 0)
+        ret = ISO_ERROR; /* should not happen after iso_node_remove_xinfo() */
+    if (ret < 0) {
+        free(md5);
+        goto ex;
+    }
+    ret = 1;
+ex:;
+    return ret;
+
+#else
+
+    return ISO_ERROR;
+
+#endif /* ! Libisofs_with_checksumS */
+
+}
+
+
+
