@@ -169,11 +169,14 @@ int read_rr_PX(struct susp_sys_user_entry *px, struct stat *st)
     st->st_nlink = iso_read_bb(px->data.PX.links, 4, NULL);
     st->st_uid = iso_read_bb(px->data.PX.uid, 4, NULL);
     st->st_gid = iso_read_bb(px->data.PX.gid, 4, NULL);
+    st->st_ino = 0;
     if (px->len_sue[0] == 44) {
         /* this corresponds to RRIP 1.12, so we have inode serial number */
         st->st_ino = iso_read_bb(px->data.PX.serial, 4, NULL);
+        /* Indicate that st_ino is valid */
+        return 2;
     }
-    return ISO_SUCCESS;
+    return 1;
 }
 
 /**
@@ -492,26 +495,14 @@ int read_aaip_AA(struct susp_sys_user_entry *sue,
      aapt = *aa_string + *aa_len;
 
      aapt[0] = 'A';
-
-#ifdef Libisofs_aaip_2_0
      aapt[1] = 'L';
-#else /* Libisofs_aaip_2_0 */
-     aapt[1] = 'A';
-#endif /* ! Libisofs_aaip_2_0 */
-
      aapt[2] = sue->len_sue[0];
      aapt[3] = 1;
      aapt[4] = 0;
-     /* Append sue payload */
 
-#ifdef Libisofs_aaip_2_0
+     /* Append sue payload */
      memcpy(aapt + 5, sue->data.AL.comps, sue->len_sue[0] - 5);
      *is_done = !(sue->data.AL.flags[0] & 1);
-#else /* Libisofs_aaip_2_0 */
-     memcpy(aapt + 5, sue->data.AA.comps, sue->len_sue[0] - 5);
-     *is_done = !(sue->data.AA.flags[0] & 1);
-#endif /* ! Libisofs_aaip_2_0 */
-
      *aa_len += sue->len_sue[0];
 
      return ISO_SUCCESS;
@@ -554,26 +545,14 @@ int read_aaip_AL(struct susp_sys_user_entry *sue,
      aapt = *aa_string + *aa_len;
 
      aapt[0] = 'A';
-
-#ifdef Libisofs_aaip_2_0
      aapt[1] = 'L';
-#else /* Libisofs_aaip_2_0 */
-     aapt[1] = 'A';
-#endif /* ! Libisofs_aaip_2_0 */
-
      aapt[2] = sue->len_sue[0];
      aapt[3] = 1;
      aapt[4] = 0;
-     /* Append sue payload */
 
-#ifdef Libisofs_aaip_2_0
+     /* Append sue payload */
      memcpy(aapt + 5, sue->data.AL.comps, sue->len_sue[0] - 5);
      *is_done = !(sue->data.AL.flags[0] & 1);
-#else /* Libisofs_aaip_2_0 */
-     memcpy(aapt + 5, sue->data.AA.comps, sue->len_sue[0] - 5);
-     *is_done = !(sue->data.AA.flags[0] & 1);
-#endif /* ! Libisofs_aaip_2_0 */
-
      *aa_len += sue->len_sue[0];
 
      return ISO_SUCCESS;
