@@ -9,6 +9,7 @@
 
 #include "util.h"
 #include "libisofs.h"
+#include "messages.h"
 #include "../version.h"
 
 #include <stdlib.h>
@@ -75,6 +76,15 @@ size_t iso_iconv(struct iso_iconv_handle *handle,
                  char **outbuf, size_t *outbytesleft, int flag)
 {
     size_t ret;
+/* The build system might indicate iconv(,const char **inbuf,) by
+   defining ICONV_CONST const
+*/
+#ifndef ICONV_CONST
+#define ICONV_CONST
+#endif
+    ICONV_CONST char **local_inbuf;
+
+    local_inbuf = (ICONV_CONST char **) inbuf;
 
     if (!(handle->status & 1)) {
         if (iso_iconv_debug)
@@ -101,7 +111,7 @@ null_buf:;
             return (size_t) -1;
         return (size_t) 0;
     }
-    ret = iconv(handle->descr, inbuf, inbytesleft, outbuf, outbytesleft);
+    ret = iconv(handle->descr, local_inbuf, inbytesleft, outbuf, outbytesleft);
     if (ret == (size_t) -1) {
         if (iso_iconv_debug)
             fprintf(stderr, "libisofs_DEBUG: iconv() failed: errno= %d %s\n",
@@ -1480,3 +1490,229 @@ int iso_init_locale(int flag)
 }
 
 
+int iso_util_encode_len_bytes(uint32_t data, char *buffer, int data_len,
+                              int *result_len, int flag)
+{
+    uint32_t x;
+    int i, l;
+    char *wpt = buffer;
+
+    if (data_len <= 0) {
+        x = data;
+        for (i = 0; i < 4 && x != 0; i++)
+            x = x >> 8;
+        l = i;
+        if (l == 0)
+            l = 1;
+    } else
+        l = data_len;
+    *((unsigned char *) (wpt++)) = l;
+    for (i = 0; i < l; i++)
+        *((unsigned char *) (wpt++)) = data >> (8 * (l - i - 1));
+    *result_len = l + 1;
+    return ISO_SUCCESS;
+}
+
+
+int iso_util_decode_len_bytes(uint32_t *data, char *buffer, int *data_len,
+                              int buffer_len, int flag)
+{
+    int i;
+
+    *data = 0;
+    *data_len = ((unsigned char *) buffer)[0];
+    if (*data_len > buffer_len - 1)
+        *data_len = buffer_len - 1;
+    for (i = 1; i <= *data_len; i++)
+        *data = (*data << 8) | ((unsigned char *) buffer)[i];
+    return ISO_SUCCESS;
+}
+
+
+int iso_util_dec_to_uint32(char *dec, uint32_t *value, int flag)
+{
+    double num;
+
+    sscanf(dec, "%lf", &num);
+    if (num < 0 || num > 4294967295.0)
+        return 0;
+    *value = num;
+    return 1;
+}
+
+
+int iso_util_hex_to_bin(char *hex, char *bin, int bin_size, int *bin_count,
+                        int flag)
+{
+    static char *allowed = {"0123456789ABCDEFabcdef"};
+    char b[3];
+    int i;
+    unsigned int u;
+
+    b[2] = 0;
+    *bin_count = 0;
+    for (i = 0; i < bin_size; i++) {
+        b[0] = hex[2 * i];
+        b[1] = hex[2 * i + 1];
+        if (strchr(allowed, b[0]) == NULL || strchr(allowed, b[1]) == NULL)
+    break;
+        sscanf(b, "%x", &u);
+        ((unsigned char *) bin)[i] = u;
+        (*bin_count)++;
+    }
+    return (*bin_count > 0);
+}
+
+
+int iso_util_tag_magic(int tag_type, char **tag_magic, int *len, int flag)
+{
+    static char *magic[] = {"",
+        "libisofs_checksum_tag_v1",
+        "libisofs_sb_checksum_tag_v1",
+        "libisofs_tree_checksum_tag_v1",
+        "libisofs_rlsb32_checksum_tag_v1"};
+    static int magic_len[]= {0, 24, 27, 29, 31};
+    static int magic_max = 4;
+
+    *tag_magic = NULL;
+    *len = 0;
+    if (tag_type < 0 || tag_type > magic_max)
+        return ISO_WRONG_ARG_VALUE;
+    *tag_magic = magic[tag_type];
+    *len = magic_len[tag_type];
+    return magic_max;
+}
+
+
+int iso_util_decode_md5_tag(char data[2048], int *tag_type, uint32_t *pos,
+                            uint32_t *range_start, uint32_t *range_size,
+                            uint32_t *next_tag, char md5[16], int flag)
+{
+    int ret, bin_count, i, mode, magic_first = 1, magic_last = 4;
+    int magic_len = 0;
+    char *cpt, self_md5[16], tag_md5[16], *tag_magic;
+    void *ctx = NULL;
+
+    *next_tag = 0;
+    mode = flag & 255;
+    if (mode > magic_last)
+        return ISO_WRONG_ARG_VALUE;
+    if (mode > 0)
+        magic_first = magic_last = mode;
+    for (i = magic_first; i <= magic_last; i++) {
+        iso_util_tag_magic(i, &tag_magic, &magic_len, 0);
+        if (strncmp(data, tag_magic, magic_len) == 0)
+    break;
+    }
+    if (i > magic_last )
+        return 0;
+    *tag_type = i;
+    cpt = data + magic_len + 1;
+    if (strncmp(cpt, "pos=", 4) != 0)
+        return 0;
+    cpt+= 4;
+    ret = iso_util_dec_to_uint32(cpt, pos, 0);
+    if (ret <= 0)
+        return 0;
+    cpt = strstr(cpt, "range_start=");
+    if (cpt == NULL)
+        return(0);
+    ret = iso_util_dec_to_uint32(cpt + 12, range_start, 0);
+    if (ret <= 0)
+        return 0;
+    cpt = strstr(cpt, "range_size=");
+    if (cpt == NULL)
+        return(0);
+    ret = iso_util_dec_to_uint32(cpt + 11, range_size, 0);
+    if (ret <= 0)
+        return 0;
+    if (*tag_type == 2 || *tag_type == 3) {
+        cpt = strstr(cpt, "next=");
+        if (cpt == NULL)
+            return(0);
+        ret = iso_util_dec_to_uint32(cpt + 5, next_tag, 0);
+        if (ret <= 0)
+            return 0;
+    } else if (*tag_type == 4) {
+        cpt = strstr(cpt, "session_start=");
+        if (cpt == NULL)
+            return(0);
+        ret = iso_util_dec_to_uint32(cpt + 14, next_tag, 0);
+        if (ret <= 0)
+            return 0;
+    }
+    cpt = strstr(cpt, "md5=");
+    if (cpt == NULL)
+        return(0);
+    ret = iso_util_hex_to_bin(cpt + 4, md5, 16, &bin_count, 0);
+    if (ret <= 0 || bin_count != 16)
+        return 0;
+
+    cpt += 4 + 32;
+    ret = iso_md5_start(&ctx);
+    if (ret < 0)
+        return ret;
+    iso_md5_compute(ctx, data , cpt - data);
+    iso_md5_end(&ctx, tag_md5);
+    cpt = strstr(cpt, "self=");
+    if (cpt == NULL)
+        return(0);
+    ret = iso_util_hex_to_bin(cpt + 5, self_md5, 16, &bin_count, 0);
+    if (ret <= 0 || bin_count != 16)
+        return 0;
+    for(i= 0; i < 16; i++)
+      if(self_md5[i] != tag_md5[i])
+        return ISO_MD5_AREA_CORRUPTED;
+    if (*(cpt + 5 + 32) != '\n')
+        return 0;
+    return(1);
+}
+
+
+int iso_util_eval_md5_tag(char *block, int desired, uint32_t lba,
+                          void *ctx, uint32_t ctx_start_lba, 
+                          int *tag_type, uint32_t *next_tag, int flag)
+{
+    int decode_ret, ret;
+    char md5[16], cloned_md5[16];
+    uint32_t pos, range_start, range_size;
+    void *cloned_ctx = NULL;
+
+    *tag_type = 0;
+    decode_ret = iso_util_decode_md5_tag(block, tag_type, &pos,
+                                  &range_start, &range_size, next_tag, md5, 0);
+    if (decode_ret != 1 && decode_ret != ISO_MD5_AREA_CORRUPTED)
+        return 0;
+    if (*tag_type > 30)
+        goto unexpected_type;
+
+    if (decode_ret == ISO_MD5_AREA_CORRUPTED) {
+        ret = decode_ret; 
+        goto ex;
+    } else if (!((1 << *tag_type) & desired)) {
+unexpected_type:;
+        iso_msg_submit(-1, ISO_MD5_TAG_UNEXPECTED, 0, NULL);
+        ret = 0;
+        goto ex;
+    } else if(pos != lba) {
+        ret = ISO_MD5_TAG_MISPLACED;
+        goto ex;
+    } else if(range_start != ctx_start_lba) {
+        ret = ISO_MD5_TAG_MISPLACED;
+    }
+    ret = iso_md5_clone(ctx, &cloned_ctx);
+    if (ret < 0)
+        goto ex;
+    iso_md5_end(&cloned_ctx, cloned_md5);
+    if (! iso_md5_match(cloned_md5, md5)) {
+        ret = ISO_MD5_TAG_MISMATCH;
+        goto ex;
+    }
+    ret = 1;
+ex:;
+    if (ret < 0)
+        iso_msg_submit(-1, ret, 0, NULL);
+    return ret;
+}
+
+ 
